<!-- Copyright 2024 Remy Blank <remy@c-space.org> -->
<!-- SPDX-License-Identifier: MIT -->

# Python

The `{exec} python` block allows executing Python code directly in the browser.

```{exec} python
:when: load
:class: hidden
import platform
print(f"{platform.python_implementation()} {'.'.join(platform.python_version_tuple())}"
      f" on {platform.platform()}")
```

## Module setup

Module setup code can be defined as a named `{exec} python` block, to be
referenced in the `:after:` option of other blocks.

```{exec} python
:name: python-setup
:when: never
def factorial(n):
  res = 1
  for i in range(2, n + 1):
    res *= i
  return res
```

## Program output

The terminal output generated by an `{exec} python` block via `sys.stdout` and
`sys.stderr` (and therefore, via `print()`) is displayed in an output block.
Output to `sys.stderr` is colored.

```{exec} python
:after: python-setup
:editable:
for i in range(10):
  print(f"factorial({i}) = {factorial(i)}")

import sys
sys.stderr.write("Program terminated.\n")
```

The terminal output block can be cleared with the form-feed control character
(`\x0c`).

```{exec} python
:when: click
:editable:
import asyncio

for i in range(10, 0, -1):
  print(f"\x0c{i}...")
  await asyncio.sleep(1)
print("\x0cHappy new year!")
```

## Program input

User input can be requested by `await`ing functions available in the global
environment. Unfortunately, `sys.stdin` (and anything that depends on it) cannot
be used, due to its blocking nature.

### Line of text

`await input_line(prompt=None)`

- `prompt`: The text to display before the input field.
- Returns the content of the input field as a `str`.

```{exec} python
:when: load
:editable:
name = await input_line("What is your name?")
print(f"Hello, {name}!")
```

### Multi-line text

`await input_text(prompt=None)`

- `prompt`: The text to display before the input field.
- Returns the content of the input field as a `str`.

```{exec} python
:when: load
:editable:
print("Please enter some text.")
text = await input_text()
print(f"\x0cThe text was:\n-------------\n{text}")
```

### Buttons

`await input_buttons(prompt, labels)`

- `prompt`: The text to display before the buttons, or `None` to not display a
  prompt.
- `labels`: The labels of the buttons, as a `list`.
- Returns the index of the button that was clicked, as an `int`.

```{exec} python
:when: load
:editable:
colors = ["Red", "Green", "Blue"]
index = await input_buttons("Pick a color:", colors)
print(f"You picked: {colors[index]}")
```

### Pause

`await pause(prompt=None, label="@icon{forward-step}")`

- `prompt`: The text to display before the button, or `None` to not display a
  prompt.
- `label`: The label of the button. If the label has the format `@icon{...}`,
  the corresponding icon from
  [Font Awesome](https://fontawesome.com/icons/categories) is used.

```{exec} python
:when: load
:editable:

n = 5
fact = 1
for i in range(2, n + 1):
  fact *= i
  await pause(f"i={i}, fact={fact}")
print(f"The factorial of {n} is {fact}")
```

## Exceptions

Exceptions that propagate out of the `{exec} python` block are displayed as a
traceback on `sys.stderr`.

```{exec} python
:after: python-setup
:editable:
def outer():
  try:
    inner()
  except Exception as e:
    raise Exception("inner() failed") from e

def inner():
  raise Exception("Something is broken")

outer()
```

## Concurrency

All `{exec} python` blocks on a page are executed in a shared, single-threaded
interpreter. Therefore, only one block can run at any given time. Nevertheless,
concurrent execution is possible through `async` coroutines. The
[`asyncio`](https://docs.python.org/3/library/asyncio.html) module provides a
lot of functionality related to `async` concurrency.

```{exec} python
:when: load
:editable:
import asyncio
import time

while True:
  print(f"\x0c{time.strftime('%Y-%m-%d %H:%M:%S')}")
  await asyncio.sleep(1)
```

```{exec} python
:when: load
:editable:
import asyncio

i = 0
while True:
  print(f"\x0ci={i}")
  i += 1
  await asyncio.sleep(0.2)
```
